#!/usr/bin/env sage -python

import sys
from sage.all import *

# This is a sage code for optimizing triangles...
# How to run it:
#  old sage (5.1) first change to correct directory and then run the file using load:
#    cd ~/Dropbox/InducedC5/bernard    # write YOUR directory :-)
#    load solve-xi.py
#
#  in sage (5.9), you can use
#    runfile solve-claim5.py

#
#
# Our main tool is http://en.wikipedia.org/wiki/Lagrange_multiplier
# We have several constraints that we need to satisfy. So what we do is that we pick which
# constraints are satisfied with equality and which are not. Then we have only equality
# constraints. For every set of equality constraints, we create L(x,\lambda) and solve it. 
# We take all the solutions and pick those that satisfy our constraints and choose the
# best one.
#
# Let us note that constraints x1 >= 0, x2 >=0, f >=0  get a special treatment.
# We are not using \lambda_{x1}... for them. If we get that x1 = 0, we just replace it with
# zero everywhere and do not add partial derivative of L(x,\lambda) by x1 = 0 (since x1
# is now a constant.
#

# names of variables
vs = var('x1 x2 x3 x4 x5 x0 f lc ls') #  lc ... lambda constraint

# function to optimize
objective=(x1)
#objective=(f)
#objective=(x0)


# extra constraints
addon=",x2==x3,x2==x4,x2==x5"


notfixedsolution = 0  # number of cases where the solution is not just one point
maxf = 0 # maximum value of f found so far in the process
maxi = 0 # i for which the maximum is reached
maxsolution = [] # contains the best solution so far
minf = 10000 # minimum value of f found so far in the process
mini = 0 # i for which the minimum is reached
minsolution = [] # contains the best solution so far

# We generate number i = 0...1023 and look at its binary representation (transformed to array).
# The array has 10 entries and each of them corresponds to one constraint being equality.
# entry 0 means free and 1 means =. The order in the field is
#
# ID 0  1   2   3   4   5   
#   lc l1l l1u l2l l2u lfl  
#
# Note that we actually do not use l1l l1u l2l l2u llf - they are l1l stands for 'lamambda x_1 lower bound',...  We rather made a substitution
# directly.


# APMonitor solution

constrcnt = 8 # number of constraints
programs =  pow(2,constrcnt)  
binformat = "{0:0"+str(constrcnt)+"b}" # makes something like "{0:012b}"


for i in range(0,programs):
	active = [ int(z) for z in list(binformat.format(i)) ] # creating {0,1} array
	
	# if any of the x_1 == 0, we are doomed anyway
	if active[1] == 1 or active[2] == 1 or active[3] == 1 or active[4] == 1 or active[5] == 1:
		continue
	
	# some quick kills.... lower and upper bounds cannot be at the same time
	# Building L, where L is the Lagrangian. Note that we ALWAYS include all parts of the Lagrangian in L
	# and when we actually use is, we may disable parts of it by setting say l1 = 0 and not taking
	# the partial derivative according to l1 in to the system of equations.
	Lstr = "L =  ("+str(objective)+" + lc*(2.0*(x1*x2+x1*x3+x1*x4+x1*x5+x2*x3+x2*x4+x2*x5+x3*x4+x3*x5+x4*x5) - 2.0*f - 3.98*(x1*x1 + x2*x2 + x3*x3 + x4*x4+x5*x5) - 0.003979) + ls*(x1+x2+x3+x4+x5+x0-1))"
	# Now we make the system of equations used in \nabla L
	# the following are always there
	eqs = "[ L.diff(ls)==0"
	# for lambdas we decide if we include them in the gradient or not. The first column
	# corresponds to not including (means not equality) and the second column means including
	if active[0] == 0:
		eqs += ",lc == 0";
	else:
		eqs += ",L.diff(lc) == 0";
	if active[1] == 0:
		eqs += ",L.diff(x1) == 0"
	else:
		eqs += ",x1 == 0"
	if active[2] == 0:
		eqs += ",L.diff(x2) == 0"
	else:
		eqs += ",x2 == 0"
	if active[3] == 0:
		eqs += ",L.diff(x3) == 0"
	else:
		eqs += ",x3 == 0"
	if active[4] == 0:
		eqs += ",L.diff(x4) == 0"
	else:
		eqs += ",x4 == 0"
	if active[5] == 0:
		eqs += ",L.diff(x5) == 0"
	else:
		eqs += ",x5 == 0"		
	if active[6] == 0:
		eqs += ",L.diff(x0) == 0"
	else:
		eqs += ",x0 == 0"
	if active[7] == 0:
		eqs += ",L.diff(f) == 0"
	else:
		eqs += ",f == 0"
	eqs += addon+"]"
	# Now we have all the equations from gradient in eqs so we create a command for solving them
	command="solution=solve("+eqs+",(x1,x2,x3,x4,x5,x0,f,lc,ls), solution_dict=True)"
	exec Lstr	# this should actually be fine do do just once at the beginning
	exec command # solving the equations - solution is in variable solution
	# Now we check how many solutions we have
	if len(solution) == 0:
		pass
		print i,active,"No solution!!"
	#if len(solution) == 1:
	#print i,active,len(solution)
	for solutionID in range(len(solution)):
		# print solutionID
		# Now we check if all constraints are satisfied. We verify constraints in the form
		#  g(x) <= 0   
		constraintsg = [ 
			-(x1), -(x2), -(x3), -(x4), -(x5), -(f), -(x0), # r2 >= 0 kind
			-(2.0*(x1*x2+x1*x3+x1*x4+x1*x5+x2*x3+x2*x4+x2*x5+x3*x4+x3*x5+x4*x5) - 2.0*f - 3.98*(x1*x1 + x2*x2 + x3*x3 + x4*x4+x5*x5) - 0.003979)
		]
		feasiblesolution = True
		for g in constraintsg:  # test all constraints
			try:
				value = g.subs(solution[solutionID]) # substituting to g
				float(value) # try if it is a number (and no some free variables)
				#print g,float(value)
				if value > 0.00001: # constraint g <= 0 violated (epsilon tolerance needed)
					print i,active,"Violates constraint ",g,"<= 0 evaluated as",float(value),"<=0"
					feasiblesolution = False
					break
			except TypeError:
				print i,active,"Not a point solution for constraint"
				print solution
				notfixedsolution = notfixedsolution + 2
				#feasiblesolution = False
				#break
				pass
		if feasiblesolution == False:
			continue # do not process the thing further if constraints violated
		# Finally, we try to evaluate f
		value = objective.subs(solution[solutionID]) # substituting to f
		try:
			#float(value)			
			#valf = float(value)
			# if there were free variables, just evaluate them as zeros...
			valf=float(value.subs({value.arguments()[x]:0 for x in range(0,len(value.arguments()))})) # evaluate it
			if maxf < valf: # test if we got a better new solution
				maxf = valf
				maxsolution=solution[solutionID]
				maxi = i
			if minf > valf: # test if we got a better new solution
				minf = valf
				minsolution=solution[solutionID]
				mini = i
			print i,active,"Got value",valf,"min is",minf,"max is",maxf
			#print solution[solutionID]
		except TypeError: # this happens if value is not a float - means not unique solution - not in claim 5
			notfixedsolution = notfixedsolution + 1
			print i,active,"Not a float!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! cnt: ", notfixedsolution,len(value.arguments())

# final wrap up	
print "Summary:"
print 'Number of not floats is',notfixedsolution
print 'mainimim is for mini=',mini,'with value minf=',minf,'and solution maxsolution=',minsolution
print 'maximum is for maxi=',maxi,'with value maxf=',maxf,'and solution maxsolution=',maxsolution


print
print 'Minimum solution: ',minf,'<=',objective
for myvar in vs:
	try:
		print myvar,'=',float(minsolution[myvar])
	except TypeError: # this happens if value is not a float
		print myvar,'=',minsolution[myvar]
		
	
print
print 'Maximum solution: ',maxf,'>=',objective
for myvar in vs:
	try:
		print myvar,'=',float(maxsolution[myvar])
	except TypeError: # this happens if value is not a float
		print myvar,'=',maxsolution[myvar]

